<?php
/*
PiAnoS (Picture Annotation System), A fingerprint examiner training tool
Copyright (C) 2011-2012 University of Lausanne, Switzerland
https://ips-labs.unil.ch/pianos

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see http://www.gnu.org/licenses/.
*/


require_once("unil/Database.php");

class ValidationError extends Exception
{
	function __construct($errors)
	{
		$message = "Validation error:\n";
		foreach ($errors as $err)
			$message .= $err['message']."\n";
		parent::__construct($message);
		$this->errors = $errors;
	}
	
	function getErrors() { return $this->errors; }
	
	private $errors;
}

class AppDB extends Database
{
	function __construct()
	{
		parent::__construct();
		$this->validation = array();
		$this->app = Application::instance();
	}
	
	// checkAdmin() *must* be called at the beginning of each function requiring
	// administration privileges. This opt-out scheme is less secure than an opt-in
	// mechanism where only non-admin functions can be called, except if a certain
	// flag is set on the function. The opt-in mechanism could be, on top of my head:
	// - have RemoteProcedureCall check in a catalog functions that are accessible
	//   to non-admin users
	//   => tedious, slow and not guaranteed to work: these methods could be called from
	//      any PHP code on the server side, defeating the check in RemoteProcedureCall
	// - "inject" a function call at the beginning of each method here.
	//   => how to do that with PHP ?
	// - split this class into 2 distinct classes: one for the users and one for the
	//   admins (possibly, a third parent class offering shared methods)
	//   => need redesign + problem when the application starts: we need to interrogate
	//      the DB to know if a user is admin or not, so we'd have to first create a 
	//      very simplistic database class that only knows how to log-in, then dispose
	//      of the connection, and create a new connection with the appropriate class
	//      (this solution is in use in e.g. PiAnoS 3.x)
	
	// So, for the time being, call $this->checkAdmin() for all functions that can be 
	// called only by admins (almost all, actually).
	
	private function checkAdmin()
	{
		if (!$_SESSION['user']['isadmin'])
			throw new Exception("Access denied");
	}
	
	// Objects management
	public function getAnyAdmin()
	{
		$res = $this->query("SELECT person.id FROM person,membership,pgroup WHERE person.id = membership.person AND membership.pgroup = pgroup.id AND pgroup.tutors IS true ORDER BY id ASC LIMIT 1;");
		if (count($res) != 1)
			throw new Exception("No admin found");
		return (int)$res[0]['id'];
	}
	
	public function setObjectOwnership($objID, $userID=NULL)
	{
		if ($userID === NULL || !$_SESSION['user']['isadmin'])
		{
			// Non-admins can only "relinquish" ownership of an object - that is, they can't
			// give it to anyone else but an admin.
			$userID = $this->getAnyAdmin();
		}
		
		$this->query("UPDATE object SET creator = ".(int)$userID." WHERE id = ".$objID.";");
	}
	
	// User management
	public function createUserStub($username, $group, $withPassword = false)
	{
		$this->checkAdmin();
		
		$nameEsc = "E'".pg_escape_string($username)."'";

		if (mb_strlen($username,'utf8') < 2)
			$this->pushValidationError('username', _('Username too short'));
		$res = $this->query("SELECT count(*) FROM person WHERE username = ".$nameEsc.";");
		if ((int)$res[0]['count'] !== 0)
			$this->pushValidationError('username', _('Username already in use'));
		$this->flushValidationErrors();

		if ($withPassword)
		{
			$pwgen = Application::get('pwgen_path');
			if (!$pwgen)
				throw new Exception("Cannot create a user with a random password: the 'pwgen_path' configuration setting has not been defined");
			
			$command = escapeshellcmd($pwgen);
			$output = '';
			$retval = 0;
			$lastline = exec($command, $output, $retval);
			
			if ($retval !== 0 || strlen($lastline) == 0)
				throw new Exception("The command '".$command."' failed. Return code: ".$retval);
				
			$password = trim($lastline);
			$passwordStored = md5($username.':'.$password);
		}
		else
		{
			$password = false;
			$passwordStored = 'will never match';
		}
		
		$this->beginTransaction();
		
		$res = $this->query("INSERT INTO person (username,name,password,language,active,suitability_approach) VALUES (".$nameEsc.",".$nameEsc.",E'".pg_escape_string($passwordStored)."',(SELECT id FROM language ORDER BY id LIMIT 1),'".($withPassword?'true':'false')."',2) RETURNING id;");
		$userID = $res[0]['id'];
		
		$this->query("INSERT INTO membership(pgroup,person) VALUES (".(int)$group.",".(int)$userID.");");
		
		$this->commitTransaction();	
		
		return array($userID,$password);
	}
	
	public function createUser($username, $name, $password, $group)
	{
		$this->checkAdmin();
		$ret = $this->createUserStub($username, $group, false);
		$uid = $ret[0];
		$this->query("UPDATE person SET name=E'".pg_escape_string($name)."',password=E'".pg_escape_string(md5($username.':'.$password))."', active=true WHERE id = ".$uid.";");
	}
	
	public function getUser($id, $withAssignments=false)
	{
		$this->checkAdmin();
		
		$res = $this->query("SELECT id,username,name,active,email FROM person WHERE id = ".(int)$id.";");
		if (count($res) != 1)
			throw new Exception("Access denied");
		
		$user = $res[0];
		$user['active'] = $user['active'] === 't';
		
		$res = $this->query("SELECT pgroup.* FROM pgroup,membership WHERE pgroup.id = membership.pgroup AND membership.person = ".(int)$id.";");
		$user['groups'] = $res;
		
		$user['tutor'] = false;
		foreach ($user['groups'] as &$g)
		{
			$g['tutors'] = $g['tutors'] === 't';
			if ($g['tutors']) $user['tutor'] = true;
		}
		
		if ($withAssignments)
		{	
			$this->checkAdmin();
			$user['assignments'] = $this->getUserAssignments($user['id']);
		}
		
		return $user;
	}

	public function saveUser($id, $name, $email, $groups)
	{
		$this->checkAdmin();
		
		require_once("lib/is_email.php");
		$email = trim($email);
		if (strlen($email) && !is_email($email))
			$this->pushValidationError('email', _('Not a valid email address'));
		$this->flushValidationErrors();
		
		$this->beginTransaction();
		$res = $this->query("UPDATE person SET name = E'".pg_escape_string($name)."', email = E'".pg_escape_string($email)."' WHERE id = ".(int)$id." RETURNING id;");
		if (count($res) != 1 || (int)$res[0]['id'] != (int)$id)
			throw new Exception("Access denied");
		$this->query("DELETE FROM membership WHERE person = ".(int)$id.";");
		foreach ($groups as $group)
			$this->query("INSERT INTO membership(person,pgroup) VALUES (".(int)$id.",".(int)$group.");");
		$this->commitTransaction();	
	}

	public function isPasswordValid($hashedPassword)
	{
		// The strength of the password can only be checked on the client side: we'll
		// never see on this side the actual password. The best we can do is check that
		// the password is not an empty string - heck, we can't even be sure it's not a
		// single space character, or a multiple of them, or the string 'pwd'.
		// Since that's awfully unuseful, we only check that the provided password is
		// a valid md5 sum.
		return preg_match('/^[a-zA-Z0-9]{32}$/', $hashedPassword);
	}
	
	public function setUserPassword($id, $newPassword)
	{
		$this->checkAdmin();
		
		if (!$this->isPasswordValid($newPassword))
			$this->pushValidationError('newPassword', _('Not a valid password'));
		$this->flushValidationErrors();
			
		$res = $this->query("UPDATE person SET password = E'".pg_escape_string($newPassword)."' WHERE id = ".(int)$id." RETURNING id;");
		if (count($res) != 1 || (int)$res[0]['id'] != (int)$id)
			throw new Exception("Access denied");
	}
	
	public function setUserLanguage($id, $langShort)
	{
		$res = $this->query("UPDATE person SET language = (SELECT id FROM language WHERE short = E'".pg_escape_string($langShort)."') WHERE id = ".(int)$id." RETURNING id;");
		if (count($res) != 1 || (int)$res[0]['id'] != (int)$id)
			throw new Exception("Access denied");
	}
	
	public function toggleUserActive($id)
	{
		$this->checkAdmin();
		
		$res = $this->query("UPDATE person SET active = NOT active WHERE id = ".(int)$id." RETURNING active;");
		if (count($res) != 1) 
			throw new Exception("Access denied");
		return $res[0]['active'] === 't';
	}
	
	public function getUsers($groupID = NULL, $showInactive = false)
	{
		$this->checkAdmin();
		
		$res = $this->query("
			SELECT 
				person.id,
				person.username,
				person.name,
				person.active,
				person.email,
				to_array(pgroup.id) AS groupids,
				to_array(pgroup.name) AS groups
			FROM person,membership,pgroup
			WHERE person.id = membership.person
				AND membership.pgroup = pgroup.id
				".($groupID ? 'AND pgroup.id = '.(int)$groupID : '')."
				".($showInactive ? '' : 'AND person.active IS true')."
			GROUP BY
				person.id,
				person.username,
				person.name,
				person.active,
				person.email
			ORDER BY person.name ASC;
		");
		foreach ($res as &$row)
		{
			$row['groupids'] = $this->intSqlArrayToPhp($row['groupids']);
			$row['groups'] = $this->stringSqlArrayToPhp($row['groups']);
			$row['active'] = $row['active'] === 't';
		}
		return $res;
	}
	
	public function deleteUser($id)
	{
		$this->checkAdmin();
		
		$this->beginTransaction();

		$this->app->callPlugins('deleteUser', $id);
		
		$res = $this->query("SELECT assignment.id FROM assignment,object WHERE assignment.id = object.id AND object.creator = ".(int)$id.";");
		foreach ($res as $row)
		{
			$this->deleteAssignment($row['id']);
		}
		
		$res = $this->query("DELETE FROM membership WHERE person = ".(int)$id.";");
		$res = $this->query("DELETE FROM security WHERE object = ".(int)$id.";");
		$res = $this->query("DELETE FROM security WHERE trustee = ".(int)$id.";");
		$res = $this->query("DELETE FROM person WHERE id = ".(int)$id.";");
		$res = $this->query("DELETE FROM trustee WHERE id = ".(int)$id.";");
		$res = $this->query("DELETE FROM object WHERE id = ".(int)$id.";");
		$this->commitTransaction();
	}
	
	public function getUserDataSummary($id)
	{
		$this->checkAdmin();
		
		$res = $this->query("SELECT count(assignment.*) FROM assignment,object WHERE assignment.id = object.id AND object.creator = ".(int)$id.";");
		
		return array('assignments'=>(int)$res[0]['count']);
	}
	
	public function updateUser($id, $name = NULL, $group = NULL, $active = NULL)
	{
		$this->checkAdmin();
		
		$token = $this->beginTransactionIf();
		
		if ($name)
		{
			$this->query("UPDATE person SET name = E'".pg_escape_string($name)."' WHERE id = ".(int)$id.";");
		}
		
		if ($group)
		{
			$this->query("DELETE FROM membership WHERE person = ".(int)$id.";");
			$this->query("INSERT INTO membership(person,pgroup) VALUES (".(int)$id.",".(int)$group.");");
		}
		
		if (is_bool($active))
		{
			$this->query("UPDATE person SET active = ".($active?'true':'false')." WHERE id = ".(int)$id.";");
		}
		
		$this->commitTransactionIf($token);
	}
	
	// Exercise management
	public function getExercise($id, $withProgress=false, $withAssignments=false)
	{
		if ($_SESSION['user']['isadmin'])
		{
			$check1  = '';
			$check2 = '';
		}
		else
		{
			$check1 = ' AND checkACL(folder.id)';
			$check2 = ' AND person.id = '.(int)$_SESSION['user']['id'];
		}
		$res = $this->query("SELECT exercise.id,exercise.name,exercise.description,exercise.mark,exercise.print,person.id AS ownerid,person.name AS ownername FROM exercise LEFT OUTER JOIN folder ON folder.id = exercise.folder LEFT OUTER JOIN person ON folder.personalof = person.id ".$check2." WHERE exercise.id = ".(int)$id." ".$check1.";");
		if (count($res) != 1)
			throw new Exception("Access denied");
		
		if ($withProgress)
			$res[0]['progress'] = $this->getExerciseProgress($id);
		
		if ($withAssignments)
		{
			$this->checkAdmin();
			$res[0]['assignments'] = $this->getExerciseAssignments($id);
		}
		
		return $res[0];
	}

	public function createExercise($name, $description, $folder, $mark, $print)
	{
		if (!$_SESSION['user']['isadmin'])
		{
			// Users can only create exercises in their personal folder
			$res = $this->query("SELECT personalof = ".(int)$_SESSION['user']['id']." AS check FROM folder WHERE id = ".(int)$folder.";");
			if (count($res) != 1 || $res[0]['check'] != 't')
				throw new Exception("Access denied");
		}
		
		$res = $this->query("INSERT INTO exercise (name, description, folder, mark, print) 
						VALUES (E'".pg_escape_string($name)."', E'".pg_escape_string($description)."',
							".(int)$folder.", ".(int)$mark.", ".(int)$print.") RETURNING id;");
		
		return (int)$res[0]['id'];
	}
	
	public function deleteExercise($id)
	{
		$this->checkAdmin();
		
		$trtoken = $this->beginTransactionIf();
		
		$res = $this->query("SELECT id FROM assignment WHERE exercise = ".(int)$id.";");
		foreach ($res as $row)
		{
			$this->deleteAssignment($row['id']);
		}

		$res = $this->query("DELETE FROM security WHERE object = ".(int)$id.";");
		$res = $this->query("DELETE FROM exercise WHERE id = ".(int)$id.";");
		$res = $this->query("DELETE FROM object WHERE id = ".(int)$id.";");
		$this->commitTransactionIf($trtoken);
	}
	
	public function getExerciseDataSummary($id)
	{
		$this->checkAdmin();
		
		$res = $this->query("SELECT count(assignment.*) FROM assignment WHERE assignment.exercise = ".(int)$id.";");
		
		return array('assignments'=>(int)$res[0]['count']);
	}
	
	// TODO: merge this function and the getUserAssignments in a single one
	public function getExerciseAssignments($id)
	{
		$this->checkAdmin();
		
		$res = $this->query("
			SELECT arv.*,
				object.createdon,
				object.updatedon,
				person.name AS person,
				person.id AS personid,
				getAssignmentStatus(arv.id) AS status
			FROM person,object,
					(SELECT id,
						ispreview,
						count(id) AS numresults,
						array_agg(step) AS ressteps,
						array_agg(type) AS restypes,
						array_agg(updatedon) AS resupdates,
						array_agg(conclusion) AS conclusions,
						array_agg(mark_value) AS mark_values,
						array_agg(decision) AS decisions,
						array_agg(decision_ex) AS decisions_ex
						FROM
							(SELECT assignment.id,
								assignment.ispreview,
								resultvex.*
							FROM assignment
								LEFT OUTER JOIN resultvex ON resultvex.assignment = assignment.id
							WHERE assignment.exercise = ".(int)$id."
							ORDER BY assignment.id,resultvex.stepid,resultvex.typeid) AS rv
						GROUP BY id,ispreview
					) AS arv
			WHERE person.id = object.creator
				AND object.id = arv.id
			ORDER BY person.name,arv.id;
		");
		
		foreach ($res as &$row)
		{
			$row['progress'] = $this->translateAssignmentStatus((int)$row['status']);
			unset($row['status']);
			$row['ispreview'] = $row['ispreview'] === 't';
			$row['numresults'] = (int)$row['numresults'];
			if ($row['numresults'] == 0)
			{
				$row['restypes'] = NULL;
				$row['ressteps'] = NULL;
				$row['resupdates'] = NULL;
				$row['conclusions'] = NULL;
				$row['mark_values'] = NULL;
				$row['decisions'] = NULL;
				$row['decisions_ex'] = NULL;
			}
			else
			{
				$row['restypes'] = $this->stringSqlArrayToPhp($row['restypes']);
				$row['ressteps'] = $this->stringSqlArrayToPhp($row['ressteps']);
				$row['resupdates'] = $this->stringSqlArrayToPhp($row['resupdates']);
				$row['conclusions'] = $this->stringSqlArrayToPhp($row['conclusions']);
				$row['mark_values'] = $this->stringSqlArrayToPhp($row['mark_values']);
				$row['decisions'] = $this->stringSqlArrayToPhp($row['decisions']);
				$row['decisions_ex'] = $this->stringSqlArrayToPhp($row['decisions_ex']);
			}
		}
			
		return $res;
	}
	
	// TODO: merge this function and the getExerciseAssignments in a single one
	public function getUserAssignments($id)
	{
		$this->checkAdmin();
		
		$res = $this->query("
			SELECT arv.*,
				object.createdon,
				object.updatedon,
				exercise.id AS exid,
				exercise.name AS exname,
				getAssignmentStatus(arv.id) AS status
			FROM exercise,object,
					(SELECT id,
						exercise,
						ispreview,
						count(id) AS numresults,
						array_agg(step) AS ressteps,
						array_agg(type) AS restypes,
						array_agg(updatedon) AS resupdates,
						array_agg(conclusion) AS conclusions,
						array_agg(mark_value) AS mark_values,
						array_agg(decision) AS decisions
						FROM
							(SELECT assignment.id,
								assignment.exercise,
								assignment.ispreview,
								resultvex.*
							FROM object,assignment
								LEFT OUTER JOIN resultvex ON resultvex.assignment = assignment.id
							WHERE object.id = assignment.id
								AND object.creator = ".(int)$id."
							ORDER BY assignment.id,resultvex.stepid,resultvex.typeid) AS rv
						GROUP BY id,exercise,ispreview
					) AS arv
			WHERE arv.exercise = exercise.id
				AND object.id = arv.id
			ORDER BY exercise.name,arv.id;
		");
		
		foreach ($res as &$row)
		{
			$row['progress'] = $this->translateAssignmentStatus((int)$row['status']);
			unset($row['status']);
			$row['ispreview'] = $row['ispreview'] === 't';
			$row['numresults'] = (int)$row['numresults'];
			if ($row['numresults'] == 0)
			{
				$row['restypes'] = NULL;
				$row['ressteps'] = NULL;
				$row['resupdates'] = NULL;
				$row['conclusions'] = NULL;
				$row['mark_values'] = NULL;
				$row['decisions'] = NULL;
			}
			else
			{
				$row['restypes'] = $this->stringSqlArrayToPhp($row['restypes']);
				$row['ressteps'] = $this->stringSqlArrayToPhp($row['ressteps']);
				$row['resupdates'] = $this->stringSqlArrayToPhp($row['resupdates']);
				$row['conclusions'] = $this->stringSqlArrayToPhp($row['conclusions']);
				$row['mark_values'] = $this->stringSqlArrayToPhp($row['mark_values']);
				$row['decisions'] = $this->stringSqlArrayToPhp($row['decisions']);
			}
		}
			
		return $res;
	}

	public function saveExercise($id, $name, $description, $mark, $print)
	{
		$this->beginTransaction();
		
		if (!$_SESSION['user']['isadmin'])
		{
			$res = $this->query("SELECT checkACL(folder.id) FROM folder,exercise WHERE exercise.id=".(int)$id." AND exercise.folder = folder.id AND folder.personalof = ".(int)$_SESSION['user']['id'].";");
			if (count($res) == 0 || $res[0]['checkacl'] != 't')
				throw new Exception("Access denied");
		}
		
		$this->query("UPDATE exercise SET name = E'".pg_escape_string($name)."', description = E'".pg_escape_string($description)."', mark = ".(int)$mark.", print = ".(int)$print." WHERE id = ".(int)$id.";");
		
		$this->commitTransaction();
	}
	
	public function deleteExercisePreviews($id)
	{
		$this->checkAdmin();

		$this->beginTransaction();
		
		$results = $this->query("SELECT result.assignment,result.exstep FROM result,assignment WHERE result.assignment=assignment.id AND assignment.ispreview IS true AND assignment.exercise = ".(int)$id);
		foreach ($results as $result)
		{
			$this->query("SELECT deleteDocResults(".(int)$result['assignment'].",".(int)$result['exstep'].");");
			$this->query("DELETE FROM result WHERE assignment = ".(int)$result['assignment']." AND exstep = ".(int)$result['exstep'].";");
			$this->query("DELETE FROM assignment WHERE id = ".(int)$result['assignment'].";");
			$this->query("DELETE FROM object WHERE id = ".(int)$result['assignment'].";");
		}
		
		$this->commitTransaction();
	}
	

	// User groups management
	public function getUserGroups($noAdmins = false, $countUsers = false, $showInactive = false)
	{
		$this->checkAdmin();
		$select = 'pgroup.id,pgroup.name';
		$from = 'pgroup';
		$group = '';
		if ($countUsers)
		{
			$select .= ',count(membership.*) AS nummembers';
			$from .= ' LEFT OUTER JOIN membership ON pgroup.id = membership.pgroup';
			$group .= 'GROUP BY pgroup.id,pgroup.name,pgroup.tutors';
		}
		$where = '';
		if ($noAdmins || ($countUsers && !$showInactive))
		{
			$wherev = array();
			if ($noAdmins)
				$wherev[] = "pgroup.tutors IS false";
			if ($countUsers && !$showInactive)
			{
				$from .= " LEFT OUTER JOIN person ON membership.person = person.id AND person.active IS true";
			}
			if (count($wherev))
				$where = 'WHERE '.implode(' AND ', $wherev);
		}
		$res = $this->query("SELECT ".$select." FROM ".$from." ".$where." ".$group." ORDER BY pgroup.name ASC;");
		return $res;
	}
	
	public function createUserGroup($name)
	{
		$this->checkAdmin();
		
		$res = $this->query("INSERT INTO pgroup (name) 
						VALUES (E'".pg_escape_string($name)."') RETURNING id;");
		return $res[0]['id'];
	}
	
	public function deleteUserGroup($id)
	{
		$this->checkAdmin();
		
		// TODO: find a better way to manage the deletion of non-empty user groups
		// (If we allow it, this is not a big problem since the users will still exist
		// but the UI will not show them)
		// For the moment, such deletion is forbidden
		$res = $this->query("SELECT count(*) FROM membership WHERE pgroup = ".(int)$id.";");
		if ((int)$res[0]['count'] !== 0)
			$this->pushValidationError('id', _('Cannot delete non-empty user group'));
		$this->flushValidationErrors();

		$this->beginTransaction();
		$res = $this->query("DELETE FROM membership WHERE pgroup = ".(int)$id.";");
		$res = $this->query("DELETE FROM security WHERE object = ".(int)$id.";");
		$res = $this->query("DELETE FROM security WHERE trustee = ".(int)$id.";");
		$res = $this->query("DELETE FROM pgroup WHERE id = ".(int)$id.";");
		$res = $this->query("DELETE FROM trustee WHERE id = ".(int)$id.";");
		$res = $this->query("DELETE FROM object WHERE id = ".(int)$id.";");
		$this->commitTransaction();
	}
	
	public function getUserGroup($id)
	{
		$this->checkAdmin();
		
		$res = $this->query("SELECT id,name,tutors FROM pgroup WHERE id = ".(int)$id.";");
		if (count($res) != 1)
			throw new Exception("Access denied");
		$pgroup = $res[0];
		$pgroup['tutors'] = $pgroup['tutors'] === 't';
		return $pgroup;
	}
	
	public function saveUserGroup($id, $name, $tutors = false)
	{
		$this->checkAdmin();
		
		$res = $this->query("UPDATE pgroup SET name = E'".pg_escape_string($name)."',tutors = ".($tutors?'true':'false')." WHERE id = ".(int)$id." RETURNING id;");
		if (count($res) != 1 || (int)$res[0]['id'] != (int)$id)
			throw new Exception("Access denied");
	}
	
	
	// Languages
	public function getLanguages()
	{
		$res = $this->query("SELECT * FROM language ORDER BY id ASC;");
		return $res;
	}

	
	// Images management
	public function addImage($type, $name, $mimetype, $info, $resolution, $width, $height, $data)
	{
		$trtoken = $this->beginTransactionIf();
		
		$res = $this->query("INSERT INTO image(type,name,mimetype,info,resolution,width,height,data) VALUES (
			(SELECT id FROM imagetype WHERE short = E'".pg_escape_string($type)."'),
			E'".pg_escape_string($name)."',
			E'".pg_escape_string($mimetype)."',
			E'".pg_escape_string($info)."',
			".(int)$resolution.",".(int)$width.",".(int)$height.",
			E'".pg_escape_bytea($data)."'
		) RETURNING id;");
		$imageID = (int)$res[0]['id'];
		
		if (!$_SESSION['user']['isadmin'])
		{
			$this->query("INSERT INTO security(object,trustee) VALUES (".$imageID.",".(int)$_SESSION['user']['id'].");");
		}
		
		$this->commitTransactionIf($trtoken);
		
		return $imageID;
	}


	// Assignments
	public function getOrCreateExercisePreview($id)
	{
		$this->checkAdmin();
		
		$res = $this->query("SELECT assignment.id FROM assignment,object WHERE assignment.id=object.id AND assignment.exercise = ".(int)$id." AND object.creator = ".$_SESSION['user']['id']." AND assignment.isPreview IS true;");
		if (count($res) > 0)
		{
			return $res[0]['id'];
		}
		else
		{
			$res = $this->query("INSERT INTO assignment(exercise,ispreview) VALUES (".(int)$id.",true) RETURNING id;");
			if (count($res) != 1)
				throw new Exception("Access denied");
			return $res[0]['id'];
		}
	}
	
	public function getOrCreateAssignment($id)
	{
		$res = $this->query("SELECT assignment.id FROM assignment,object WHERE assignment.id=object.id AND assignment.exercise = ".(int)$id." AND object.creator = ".$_SESSION['user']['id']." AND assignment.isPreview IS false;");
		if (count($res) > 0)
		{
			return (int)$res[0]['id'];
		}
		else
		{
			$res = $this->query("INSERT INTO assignment(exercise,ispreview) VALUES (".(int)$id.",".($_SESSION['user']['isadmin']?'true':'false').") RETURNING id;");
			if (count($res) != 1)
				throw new Exception("Access denied");
			return (int)$res[0]['id'];
		}
	}
	
	public function getAssignment($id)
	{
		$res = $this->query("SELECT assignment.* FROM assignment,object WHERE assignment.id=object.id AND assignment.id = ".(int)$id.";");
		if (count($res) != 1)
			throw new Exception("Access denied");
		return $res[0];
	}
	
	public function deleteAssignment($aid)
	{
		$this->checkAdmin();
	
		$token = $this->beginTransactionIf();
		
		$res = $this->query("DELETE FROM minutia WHERE assignment = ".(int)$aid.";");
		$res = $this->query("DELETE FROM observation WHERE assignment = ".(int)$aid.";");
		$res = $this->query("DELETE FROM annotation WHERE assignment = ".(int)$aid.";");
		$res = $this->query("DELETE FROM result WHERE assignment = ".(int)$aid.";");
		$res = $this->query("DELETE FROM resulthistory WHERE assignment = ".(int)$aid.";");
		$res = $this->query("DELETE FROM assignment WHERE id = ".(int)$aid.";");
		$res = $this->query("DELETE FROM object WHERE id = ".(int)$aid.";");
		
		$this->commitTransactionIf($token);
	}
	
	// Folders
	public function getFolders($countExercises = false, $showPersonal = false)
	{
		$select = 'folder.id,folder.name';
		$from = 'folder';
		$group = '';
		$where = '';
		if ($countExercises)
		{
			$select .= ',count(exercise.*) AS numexercises';
			$from .= ' LEFT OUTER JOIN exercise ON folder.id = exercise.folder';
			$group .= 'GROUP BY folder.id,folder.name';
		}
		if (!$_SESSION['user']['isadmin'])
		{
			$select .= ",folder.personalof=".(int)$_SESSION['user']['id'].' AS own';
			$where = "WHERE checkACL(folder.id,".$_SESSION['user']['id'].")";
			if ($group) $group .= ',own';
		}
		else if (!$showPersonal)
		{
			$where = "WHERE personalof IS NULL";
		}
		
		$res = $this->query("SELECT ".$select." FROM ".$from." ".$where." ".$group." ORDER BY folder.name ASC;");

		if (!$_SESSION['user']['isadmin'])
		{
			foreach ($res as &$row)
			{
				$row['own'] = $row['own'] == 't';
				if ($row['own'])
					$row['name'] = 'My personal folder';
			}
		}

		return $res;
	}
	
	public function getFolderExercises($folderID = NULL, $progress = false)
	{
		$res = $this->query("
			SELECT exercise.id,exercise.name,folder.name AS folder 
			FROM exercise,folder
			WHERE exercise.folder = folder.id
			".($_SESSION['user']['isadmin']?"":"AND checkACL(folder.id,".(int)$_SESSION['user']['id'].")")."
			".($folderID ? 'AND folder.id = '.(int)$folderID : '')."
			ORDER BY exercise.name ASC;
		");
		
		if ($progress)
		{
			foreach ($res as &$ex)
				$ex['progress'] = $this->getExerciseProgress($ex['id']);
		}
		
		return $res;
	}

	public function getFolder($id)
	{
		$res = $this->query("SELECT folder.*,object.* FROM folder,object WHERE folder.id=object.id AND checkACL(folder.id) AND folder.id = ".(int)$id.";");
		if (count($res) != 1)
			throw new Exception("Access denied");
		$folder = $res[0];
		
		$res = $this->query("SELECT exercise.*,object.* FROM exercise,object WHERE exercise.id=object.id AND exercise.folder = ".(int)$id.";");
		$folder['exercises'] = $res;
		
		if ($_SESSION['user']['isadmin'])
		{
			$res = $this->query("SELECT security.trustee FROM security,pgroup WHERE security.trustee = pgroup.id AND security.object = ".(int)$id.";");
			$folder['acl'] = array();
			foreach ($res as $pgroup)
				$folder['acl'][] = (int)$pgroup['trustee'];
		}
		
		return $folder;
	}
	
	public function createFolder($name)
	{
		$this->checkAdmin();
		
		$res = $this->query("INSERT INTO folder (name) 
						VALUES (E'".pg_escape_string($name)."') RETURNING id;");
		return $res[0]['id'];
	}
	
	public function deleteFolder($id)
	{
		$this->checkAdmin();
		
		$this->beginTransaction();
		
		$res = $this->query("SELECT id FROM exercise WHERE folder = ".(int)$id.";");
		foreach ($res as $row)
			$this->deleteExercise((int)$row['id']);
		
		$res = $this->query("DELETE FROM security WHERE object = ".(int)$id.";");
		$res = $this->query("DELETE FROM folder WHERE id = ".(int)$id.";");
		$res = $this->query("DELETE FROM object WHERE id = ".(int)$id.";");

		$this->commitTransaction();
	}
	
	public function getFolderDataSummary($id)
	{
		$this->checkAdmin();
		
		$res = $this->query("SELECT exercise.id,exercise.name,count(assignment.*) AS numassignments FROM exercise LEFT OUTER JOIN assignment ON exercise.id = assignment.exercise WHERE folder = ".(int)$id." GROUP BY exercise.id,exercise.name;");
		
		$numTotalAssignments = 0;
		foreach ($res as $row)
			$numTotalAssignments += (int)$row['numassignments'];
			
		return array('assignments'=>$numTotalAssignments, 'exercises'=>count($res), 'all'=>$res);
	}

	public function saveFolder($id, $name, $groups = array())
	{
		$this->checkAdmin();
		
		$this->beginTransaction();
		$res = $this->query("UPDATE folder SET name = E'".pg_escape_string($name)."' WHERE id = ".(int)$id." RETURNING id;");
		if (count($res) != 1) 
			throw new Exception("Access denied");
		$this->query("DELETE FROM security USING pgroup WHERE security.object = ".(int)$id." AND security.trustee = pgroup.id AND pgroup.tutors IS false;");
		foreach ($groups as $group)
			$this->query("INSERT INTO security(object,trustee) VALUES (".(int)$id.",(SELECT id FROM pgroup WHERE id = ".(int)$group." AND tutors IS false));"); // The inner SELECT makes sure we are inserting only regular groups here, not users nor tutor groups.
		$this->commitTransaction();
		return $id;
	}
	
	public function setFolderACL($id, $acl)
	{
		$this->checkAdmin();
		
		$this->beginTransaction();
		$this->query("DELETE FROM security WHERE object = ".(int)$id." AND trustee IN (SELECT id FROM trustee WHERE isgroup IS true);");
		foreach ($acl as $ace)
			$this->query("INSERT INTO security(object,trustee) VALUES (".(int)$id.",".(int)$ace.");");
		$this->commitTransaction();
	}
	
	// Images
	public function getImageInfo($id)
	{
		$res = $this->query("SELECT image.id, imagetype.short AS type, image.name, image.mimetype, 
									image.info, image.resolution, image.width, image.height, image.thumb,
									(SELECT count(*) FROM image AS image2 WHERE image2.thumb = image.id) > 0 AS isThumb
			FROM image,imagetype
			WHERE image.type = imagetype.id
			AND image.id = ".(int)$id.";"
		);
		
		if (count($res) != 1)
			throw new Exception("Access denied");

		$res[0]['isthumb'] = $res[0]['isthumb'] === 't';
		
		return $res[0];
	}
	
	public function getImageLibrary($filter = NULL, $type = NULL, $extended = false)
	{
		if ($type)
			$where = "imagev.typeid = (SELECT id FROM imagetype WHERE short = E'".pg_escape_string($type)."')";
		else
			$where = "imagev.typeid <> (SELECT id FROM imagetype WHERE short = E'THUMBNAIL')";
			
		if ($filter)
			$where .= " AND imagev.name LIKE '%".pg_escape_string($filter)."%'";
		
		if (!$_SESSION['user']['isadmin'])
			$where .= " AND checkImageReadACL(imagev.id,".(int)$_SESSION['user']['id'].")";
			
		$columns = "imagev.id,imagev.typeid,imagev.type,imagev.name,imagev.mimetype,imagev.info,imagev.resolution,imagev.width,imagev.height";
		if ($extended)
		{
			$select = $columns.",person.name AS owner,object.createdon,array_agg(mex.id) AS markforexid,array_agg(pex.id) AS printforexid,
				array_agg(mex.name) AS markforex,array_agg(pex.name) AS printforex";
			$from = "person,object,imagev LEFT OUTER JOIN exercise AS mex ON mex.mark = imagev.id
					LEFT OUTER JOIN exercise AS pex ON pex.print = imagev.id";
			$where .= " AND person.id = object.creator AND object.id = imagev.id";
			$group = "GROUP BY ".$columns.",owner,createdon";
		}
		else
		{
			$select = $columns;
			$group = "";
			$from = "imagev";
		}
		$sql = 	"SELECT ".$select." FROM ".$from." WHERE ".$where." ".$group." ORDER BY name ASC;";	
		$res = $this->query($sql);
		
		if ($extended)
		{
			foreach ($res as &$row)
			{
				$row['markforexid'] = $row['markforex'] === '{NULL}' ? array() : $this->intSqlArrayToPhp($row['markforex']);
				$row['printforexid'] = $row['printforex'] === '{NULL}' ? array() : $this->intSqlArrayToPhp($row['printforex']);
				$row['markforex'] = $row['markforex'] === '{NULL}' ? array() : $this->stringSqlArrayToPhp($row['markforex']);
				$row['printforex'] = $row['printforex'] === '{NULL}' ? array() : $this->stringSqlArrayToPhp($row['printforex']);
			}
		}

		return $res;
	}
	
	public function deleteImage($id, $cascade=false)
	{
		$this->checkAdmin();
		
		$trtoken = $this->beginTransactionIf();
		
		$res = $this->query("SELECT id FROM exercise WHERE mark = ".(int)$id." OR print = ".(int)$id." GROUP BY id;");
		if (count($res) !== 0 && !$cascade)
			throw new Exception("Access denied");
		
		foreach ($res as $ex)
		{
			$this->deleteExercise($ex['id']);
		}
		
		$res = $this->query("DELETE FROM security WHERE object = ".(int)$id.";");
		$res = $this->query("DELETE FROM image WHERE id = ".(int)$id.";");
		$res = $this->query("DELETE FROM object WHERE id = ".(int)$id.";");
		
		$this->commitTransactionIf($trtoken);
	}
	
	// Results
	public function getExerciseProgress($id)
	{
		$res = $this->query("SELECT assignment.id 
			FROM assignment,object 
			WHERE object.creator = ".(int)$_SESSION['user']['id']." 
			AND object.id = assignment.id 
			AND assignment.exercise = ".(int)$id." 
			ORDER BY object.createdon 
			DESC LIMIT 1;");
		if (count($res) != 1)
			return array('status'=>-1, 'label'=>'Not started');
		else
			return $this->getAssignmentProgress($res[0]['id']);
	}
	
	public function getAssignmentProgress($aid)
	{
		try {
			$res = $this->query("SELECT getAssignmentStatus(".(int)$aid.");");
		} catch (Exception $ex) {
			throw new Exception("Access denied");
		}
		$status = (int)$res[0]['getassignmentstatus'];
		return $this->translateAssignmentStatus($status);
	}
	
	private function translateAssignmentStatus($status)
	{
		switch ($status)
		{
			default:
			case 0: $label = 'Exercise started'; $step = 'ANALYSIS'; break;
			case 1: $label = 'Analysis started'; $step = 'ANALYSIS'; break;
			case 2: $label = 'Analysis submitted'; $step = 'ANALYSIS'; break;
			case 3: $label = 'Comparison started'; $step = 'COMPARISON'; break;
			case 4: $label = 'Comparison submitted'; $step = 'COMPARISON'; break;
			case 5: $label = 'Evaluation started'; $step = 'EVALUATION'; break;
			case 6: $label = 'Evaluation submitted'; $step = 'EVALUATION'; break;
		}
		return array('status'=>$status,'label'=>$label,'step'=>$step);
	}
	
	public function getAssignmentSubmissionTime($aid, $step)
	{
		if (is_string($step))
			$subq = "(SELECT id FROM resultstep WHERE short = E'".pg_escape_string($step)."')";
		else
			$subq = (int)$step;
			
		$query = "SELECT updatedon 
			FROM result
			WHERE assignment = ".(int)$aid." 
			AND type = (SELECT id FROM resulttype WHERE short = 'VALIDATED')
			AND step = ".$subq.";";
		
		$res = $this->query($query);
		if (count($res) != 1)
			throw new Exception("Database integrity error");
		return $res[0]['updatedon'];
	}
	
	private function getResultTypes()
	{
		$res = $this->query("SELECT * FROM resulttype ORDER BY id ASC;");
		$ret = array();
		foreach ($res as $item)
			$ret[$item['short']] = (int)$item['id'];
		return $ret;
	}
	
	private function getResultSteps()
	{
		$res = $this->query("SELECT * FROM resultstep ORDER BY id ASC;");
		$ret = array();
		foreach ($res as $item)
			$ret[$item['short']] = (int)$item['id'];
		return $ret;
	}

	private function getNextResultStep($step)
	{
		if (is_string($step))
			$res = $this->query("SELECT * FROM resultstep WHERE id = (SELECT id FROM resultstep WHERE short = E'".pg_escape_string($step)."') + 1;");
		else
			$res = $this->query("SELECT * FROM resultstep WHERE id = ".((int)$step+1).";");
		if (count($res) == 1)
			return $res[0];
		else
			return NULL;
	}
	
	public function saveResults($aid, $type, $step, $mark = NULL, $print = NULL, $observations = NULL)
	{
		$rtypes = $this->getResultTypes();
		$rsteps = $this->getResultSteps();
		
		if (!array_key_exists($type,$rtypes))
			throw new Exception("Invalid parameters");
		$rtype = $rtypes[$type];
		
		if (!array_key_exists($step,$rsteps))
			throw new Exception("Invalid parameters");
		$rstep = $rsteps[$step];
		
		$this->beginTransaction();
		
		// First thing to check: whether the user has access to the assignment
		$owner = $this->query("SELECT creator FROM object WHERE id = ".(int)$aid.";");
		if (count($owner) !== 1 || (int)$owner[0]['creator'] !== $_SESSION['user']['id'])
			throw new Exception("Access denied");

		
		$data = array();
		if ($mark) $data['mark'] = $mark;
		if ($print) $data['print'] = $print;
		if ($observations) $data['observations'] = $observations;
		
		$serialized = json_encode($data);
		$e_serialized = pg_escape_string($serialized);
		
		$results = array();
		
		$res = $this->query("SELECT count(*) FROM result WHERE assignment = ".(int)$aid." AND type = ".$rtype." AND step = ".$rstep.";");
		if ((int)$res[0]['count'] !== 1)
		{
			$res = $this->query("INSERT INTO result(assignment,type,step,data) VALUES (".(int)$aid.",".$rtype.",".$rstep.",E'".$e_serialized."');");
			$results[] = array((int)$aid,$rtype,$rstep,$data);
			
			if ($type == 'VALIDATED')
			{
				// Make a copy of the result set and store it as the initial state for the next step
				// (but only do it if it makes sense - i.e. if there actually is a next step)
				$nextStep = $this->getNextResultStep($rstep);
				if ($nextStep !== NULL)
				{
					// Strip the observations from the copy - as observations during one step are
					// not supposed to be repeated in the next one.
					$dataCopy = array();
					if ($mark) $dataCopy['mark'] = $mark;
					if ($print) $dataCopy['print'] = $print;
					$dataCopyEncoded = json_encode($dataCopy);
					
					$res = $this->query("INSERT INTO result(assignment,type,step,data) VALUES (".(int)$aid.",".$rtypes['CURRENT'].",".(int)$nextStep['id'].",E'".pg_escape_string($dataCopyEncoded)."');");
					$results[] = array((int)$aid,$rtypes['CURRENT'],(int)$nextStep['id'],$dataCopy);
				}
			}
		}
		else
		{
			// Second check: one cannot update a VALIDATED result (this check is also enforced at the DB level, 
			// but the "Access denied" message seems more appropriate for the end user - who should never see
			// it unless she is fiddling with the API).
			if ($type == 'VALIDATED')
				throw new Exception("Access denied");
			
			$this->query("UPDATE result SET data = E'".$e_serialized."', updatedon = now() WHERE assignment = ".$aid." AND type = ".$rtype." AND step = ".$rstep.";");
			$results[] = array((int)$aid,$rtype,$rstep,$data);
		}
		
		// The result set itself is a JSON object, whose format is loosely defined. That is, it *must* contain
		// a number of informations, and it *may* contain more info, depending on the various plugins that are
		// running on the client side (e.g. additionnal tools that may provide additionnal features). PiAnoS is
		// designed to make at least accessible to the SQL queries the following data:
		// - all annotations (mandatory, not NULL): ID, type ('minutia', 'ridge', ...), date created, date modified
		// - all annotations (optional, may be NULL): pairing ID, subtype (e.g. 'quality', 'wcs', ...), value (arbitrary value whose interpretation depends on the subtype)
		// - minutiae: X,Y,Theta,Type
		// - ridges/areas: type, list of points, color, opacity, width
		$this->updateResultsData($results);
		
		// Archive the data in the history (the history does not have a PK, hence all results are
		// continuously added to it). Remember to purge it some day !
		$this->query("INSERT INTO resulthistory(assignment,type,step,data) VALUES (".(int)$aid.",".$rtype.",".$rstep.",E'".$e_serialized."');");

		$this->commitTransaction();
	}
	
	private function getMinutiaTypes()
	{
		$res = $this->query("SELECT * FROM minutiatype ORDER BY id ASC;");
		$ret = array();
		foreach ($res as $item)
			$ret[$item['name']] = (int)$item['id'];
		return $ret;
	}

	private function updateResultsData($results)
	{
		foreach ($results as $result)
		{
			$aid = $result[0];
			$rtype = $result[1];
			$rstep = $result[2];
			$data = $result[3];
			
			$this->query("DELETE FROM minutia WHERE assignment = ".$aid." AND restype = ".$rtype." AND resstep = ".$rstep.";");
			$this->query("DELETE FROM annotation WHERE assignment = ".$aid." AND restype = ".$rtype." AND resstep = ".$rstep.";");
			$this->query("DELETE FROM observation WHERE assignment = ".$aid." AND restype = ".$rtype." AND resstep = ".$rstep.";");
			
			if (array_key_exists('observations', $data))
			{
				foreach ($data['observations'] as $obsType => $obsValue)
				{
					$this->query("INSERT INTO observation(assignment,restype,resstep,type,value) VALUES (".$aid.",".$rtype.",".$rstep.",(SELECT id FROM observationtype WHERE name = E'".pg_escape_string($obsType)."'),E'".pg_escape_string(json_encode($obsValue))."');");
				}
			}
			
			$imgtypes = array();
			if (array_key_exists('mark', $data))
				$imgtypes['MARK'] = $data['mark'];
			if (array_key_exists('print', $data))
				$imgtypes['PRINT'] = $data['print'];
			
			foreach ($imgtypes as $imgTypeShort => $imgData)
			{
				$res = $this->query("SELECT id FROM imagetype WHERE short = E'".pg_escape_string($imgTypeShort)."';");
				$imgTypeID = (int)$res[0]['id'];
				
				$pkfields = array('assignment','restype','resstep');
				$pkvalues = array($aid, $rtype, $rstep);
				
				foreach ($imgData as $aTypeNamePlural => $annotations)
				{
					if ($aTypeNamePlural == 'image')
						continue;
					
					$basefields = array_merge($pkfields, array('image'));
					$basevalues = array_merge($pkvalues, array($imgTypeID));
					
					// FIXME!
					// This is *very* ugly: the front-end sends the annotations as arrays depending on the type
					// i.e. 'minutiae' => array_of_minutiae, 'ridges' => array_of_ridges, etc.
					// Each annotation does *not* include the actual type - which can be inferred from the name
					// of the key (e.g. things in the 'minutiae' array have the type 'minutia').
					// This is heavily broken because extensibility is basically impossible. It is also coupled
					// with the problem of finding the type of an annotation in the Javascript code - the allAnnotations
					// array vs. the annotations.minutiae, annotations.ridges, etc. thing.
					// This needs refactoring. In the meantime, we infer the type from the plural term. Additionnal
					// annotation types will not be recognized - not too much trouble for the moment.
					switch ($aTypeNamePlural) {
						case 'minutiae': $aTypeName = 'minutia'; break;
						case 'ridges': $aTypeName = 'ridge'; break;
						case 'areas': $aTypeName = 'area'; break;
						case 'comments': $aTypeName = 'comment'; break;
						default: $aTypeName = $aTypeNamePlural;
					}
					$res = $this->query("SELECT getAnnotationType(E'".pg_escape_string($aTypeName)."') AS id;");
					$aTypeID = (int)$res[0]['id'];

					foreach ($annotations as $id => $values)
					{
						$afields = array('id','type','createdon','updatedon');
						$avalues = array((int)$id, $aTypeID, "E'".pg_escape_string($values['created'])."'", "E'".pg_escape_string($values['modified'])."'");
					
						if (array_key_exists('pairing', $values))
						{
							$afields[] = 'pairing';
							$avalues[] = (int)$values['pairing'];
						}
						if (array_key_exists('type', $values) && $aTypeName != 'minutia')
						{
							// Don't write the subtype for minutiae ('bifurcation', etc. is normalized through 
							// the minutiatype table and the 'type' field of the minutia table
							$afields[] = 'subtype';
							$avalues[] = "E'".pg_escape_string($values['type'])."'";
						}
						if (array_key_exists('value', $values))
						{
							$afields[] = 'value';
							$avalues[] = (int)$values['value'];
						}
						
						$sqlfields = implode(',', array_merge($basefields, $afields));
						$sqlvalues = implode(',', array_merge($basevalues, $avalues));
						
						$this->query("INSERT INTO annotation(".$sqlfields.") VALUES (".$sqlvalues.");");
					}
				}
				
				if (array_key_exists('minutiae', $imgData))
				{
					$mtypes = $this->getMinutiaTypes();
					foreach ($imgData['minutiae'] as $id => $values)
					{
						$mfields = array('id','x','y','theta','type');
						$mvalues = array((int)$id, (double)$values['pos'][0], (double)$values['pos'][1], (double)$values['angle'], $mtypes[$values['type']]);
					
						$sqlfields = implode(',', array_merge($pkfields, $mfields));
						$sqlvalues = implode(',', array_merge($pkvalues, $mvalues));
						
						$this->query("INSERT INTO minutia(".$sqlfields.") VALUES (".$sqlvalues.");");
					}
				}
			}
		}
	}
	
	public function updateAllResults($numItems = 100)
	{
		if (!_DEBUG) throw new Exception("Invalid parameters");
		
		$this->checkAdmin();
		
		$this->beginTransaction();
		
		$alla = $this->query("SELECT assignment.id,count(result.*) AS numres FROM assignment,result WHERE result.assignment = assignment.id GROUP BY assignment.id HAVING count(result.*) > 0;");
		$ret = array();
		$itemCount = 0;
		foreach ($alla as $arow)
		{
			$aid = (int)$arow['id'];
			$res = $this->query("SELECT result.type,result.step,count(annotation.*) AS numannot,count(observation.*) AS numobs
			FROM result
				LEFT OUTER JOIN annotation ON result.assignment = annotation.assignment 
					AND result.step = annotation.resstep 
					AND result.type = annotation.restype
				LEFT OUTER JOIN observation ON result.assignment = observation.assignment 
					AND result.step = observation.resstep 
					AND result.type = observation.restype
			WHERE result.assignment = ".$aid."
			GROUP BY result.type,result.step;");
			foreach ($res as $row)
			{
				$numAnnot = (int)$row['numannot'];
				$numObs = (int)$row['numobs'];
				if ($numAnnot === 0 && $numObs === 0)
				{
					$step = (int)$row['step'];
					$type = (int)$row['type'];
					$datares = $this->query("SELECT data FROM result WHERE assignment = ".$aid." AND step = ".$step." AND type = ".$type.";");
					if (count($datares) !== 1)
						throw new Exception("Assertion failed");
					$data = json_decode($datares[0]['data'], true);
					$this->updateResultsData(array(array($aid,$type,$step,$data)));
					$ret[] = array($aid,$type,$step);
					if ($itemCount++ >= $numItems)
						break;
				}
			}
			if ($itemCount >= $numItems)
				break;
		}
		
		$this->commitTransaction();
		
		return $ret;
	}

/*	public function hasResults($aid, $exstep)
	{
		$results = $this->query("SELECT * FROM result WHERE assignment = ".(int)$aid." AND exstep = ".(int)$exstep.";");
		return count($results) > 0;
	}*/
	
	public function getResults($aid, $type, $step)
	{
		// First thing to check: whether the user has access to the assignment
		if (!$_SESSION['user']['isadmin'])
		{
			$owner = $this->query("SELECT creator FROM object WHERE id = ".(int)$aid.";");
			if (count($owner) !== 1 || (int)$owner[0]['creator'] !== $_SESSION['user']['id'])
				throw new Exception("Access denied");
		}

		$rtypes = $this->getResultTypes();
		$rtypes['LATEST'] = -1;
		$rsteps = $this->getResultSteps();
		
		if (!array_key_exists($type,$rtypes))
			throw new Exception("Invalid parameters");
		
		if (!array_key_exists($step,$rsteps))
			throw new Exception("Invalid parameters");
		$rstep = $rsteps[$step];

		if ($_SESSION['user']['isadmin'] && (int)$aid<0)
		{
			// If we're logged as admin, it is valid to receive a negative number as assignment ID. 
			// It means "all assignments for the exercise -$aid"
			$needMany = true;
			$assignments = $this->query("
			SELECT id,userid,username,getAssignmentStatus(id) AS status FROM (
				SELECT object.id, person.id AS userid, person.name AS username
				FROM assignment, object, person
				WHERE assignment.id = object.id
					AND object.creator = person.id
					AND assignment.ispreview IS false 
					AND assignment.exercise = ".-(int)$aid." 
				GROUP BY object.id, object.updatedon, person.id, person.name
				ORDER BY object.updatedon DESC
			) AS foo
			ORDER BY username ASC;");
			// We'll need this value later on
			$analysisStepID = $rsteps['ANALYSIS'];
		}
		else
		{
			$needMany = false;
			$assignments = array(array('id'=>$aid));
		}
		
		$ret = array();
		foreach ($assignments as $item)
		{
			$rtype = NULL;
			if ($type == 'LATEST')
			{
				$res = $this->query("
					SELECT result.*,resulttype.short,object.createdon 
					FROM result,resulttype,object
					WHERE result.assignment = ".(int)$item['id']."
						AND result.type = resulttype.id
						AND result.step = ".$rstep."
						AND object.id = result.assignment
					ORDER BY result.updatedon DESC, object.createdon DESC
					LIMIT 1;");
				if (count($res) == 0)
				{
					$rtype = NULL; // will fail later on
					$type = 'CURRENT';
				}
				else
					$rtype = (int)$res[0]['type'];
			}
			
			if ($rtype === NULL)
			{
				$rtype = (int)$rtypes[$type];
			}
			
			$res = $this->query("SELECT * FROM result WHERE assignment = ".(int)$item['id']." AND type = ".$rtype." AND step = ".$rstep.";");
			if (count($res) !== 1)
			{
				// No results at all
				$data = array();
			}
			else
			{
				$result = $res[0];
				$data = json_decode($result['data'], true);
			}
			
			if (!$needMany)
				return $data;
			
			// Attach additionnal data relative to the user (the annotation themselves contain the userid
			// but not the name)
			$data['user'] = array('id'=>(int)$item['userid'], 'name'=>$item['username']);
			
			// Data for the correct display of "yellow" minutiae
			$data['progress'] = array('status'=>$item['status']);
			if ($rstep > $analysisStepID && $item['status'] >= 2)
				$data['progress']['analysis_submitted_on'] = $this->getAssignmentSubmissionTime($item['id'], $analysisStepID);
			
			$ret[] = $data;
		}
		
		return $ret;
	}
	
	public function deleteAllResults($aid)
	{
		$this->checkAdmin();
		
		$results = $this->query("SELECT * FROM result WHERE assignment = ".(int)$aid.";");
//		foreach ($reults as $result)
	}
	
	// Statistics
	public function search($what)
	{
		$this->checkAdmin();
		if (strlen($what) == 0)
			throw new Exception("Invalid parameters");
		$res = $this->query("
		SELECT ts_rank(ftsv.fts, query) AS rank, ftsv.type, ftsv.id, ftsv.name, ftsv.desc
		FROM ftsv, plainto_tsquery(E'".pg_escape_string($what)."%') AS query 
		WHERE ftsv.fts @@ query
		ORDER BY rank DESC;");
		return $res;
	}
	
	public function getAllConclusions($exid, $step=NULL, $type=NULL)
	{
		$this->checkAdmin();
		
		return $this->getExerciseAssignments($exid);
		
		$columns = array('object.creator AS uid', 'observationtype.name', 'observation.value');
		$from = array('object', 'assignment', 'observationtype', 'observation');
		$where = "object.id = assignment.id
				AND assignment.exercise = ".(int)$exid."
				AND assignment.id = observation.assignment
				AND observation.type = observationtype.id";

		if ($step)
		{
			if (is_string($step))
			{
				$where .= " AND resultstep.id = observation.resstep AND resultstep.short ILIKE E'".pg_escape_string($step)."'";
				$from[] = 'resultstep';
			}
			else
				$where .= " AND observation.resstep = ".(int)$step;
				
			$columns[] = 'observation.resstep AS step';
		}
				
		if ($type)
		{
			if (is_string($type))
				$where .= " AND observationtype.name = E'".pg_escape_string($type)."'";
			else
				$where .= " AND observation.type = ".(int)$type;
		}
		
		$sql = "SELECT ".implode($columns, ',')."
			FROM ".implode($from, ', ')."
			WHERE ".$where.";";
		
		throw new Exception($sql);
		$res = $this->query($sql);
		return $res;
				
	}
	
	// User views
	public function getHomeData()
	{
		$res = $this->query("SELECT folder.id, 'My personal folder' AS name,array_agg(ex.id) AS exids,array_agg(ex.name) AS exnames FROM folder LEFT OUTER JOIN (SELECT id,name,folder FROM exercise ORDER BY name ASC) AS ex ON ex.folder=folder.id WHERE folder.personalof = ".(int)$_SESSION['user']['id']." GROUP BY folder.id;");
		if (count($res) > 1)
			throw new Exception("Database integrity error");
		
		if (count($res))
		{
			$folder = $res[0];
			$folder['exids'] = $folder['exids'] == '{NULL}' ? array() : $this->intSqlArrayToPhp($folder['exids']);
			$folder['exnames'] = $folder['exnames'] == '{NULL}' ? array() : $this->stringSqlArrayToPhp($folder['exnames']);
			$folder['name'] = 'My personal folder';
		}
		else
			$folder = NULL;
			
		$ret = array(
			'progress' => $this->getFolderExercises(NULL, true),
			'folder' => $folder,
		);
		return $ret;
	}
	
	public function createPersonalFolder($id=NULL)
	{
		$uid = $_SESSION['user']['isadmin'] ? (int)$id : (int)$_SESSION['user']['id'];
		$this->beginTransaction();
		$res = $this->query("INSERT INTO folder (name,personalof) VALUES ((SELECT name FROM person WHERE id = ".$uid.")||E'\'s folder',".(int)$_SESSION['user']['id'].") RETURNING id;");
		$fid = (int)$res[0]['id'];
		$this->query("SELECT applyACL(".$fid.",ARRAY[]::integer[], ARRAY[".$uid."]);");
		$this->commitTransaction();
	}
	
	public function getSuitabilityApproach($userid = NULL)
	{
		if ($_SESSION['user']['isadmin'])
		{
			$userid = $userid ? (int)$userid : $_SESSION['user']['id'];
		}
		else
		{
			$userid = (int)$_SESSION['user']['id'];
		}
		
		$res = $this->query("SELECT suitability_approach AS approach FROM person WHERE id = ".$userid.";");
		if (count($res) != 1)
			throw new Exception("Access denied");
		$val = (int)$res[0]['approach'];
		if ($val != 1 && $val != 2)
			throw new Exception("Database integrity error");
		return $val;
	}
	
	public function isMarkOfNoValue($assignmentID)
	{
		if (!$_SESSION['user']['isadmin'])
		{
			$res = $this->query("SELECT creator FROM object WHERE id = ".(int)$assignmentID.";");
			if ((int)$res[0]['creator'] !== (int)$_SESSION['user']['id'])
				throw new Exception("Access denied");
		}
		$res = $this->query("SELECT value FROM observation WHERE assignment = ".(int)$assignmentID." AND restype = (SELECT ID FROM resulttype WHERE short = 'VALIDATED') AND resstep = (SELECT ID FROM resultstep WHERE short = 'ANALYSIS') AND type = (SELECT ID FROM observationtype WHERE name = 'mark-value');");
		if (count($res) != 1)
			throw new Exception("Database integrity error");
		$val = (int)$res[0]['value'];
		if ($val !== 0 && $val !== 1 && $val !== 2)
			throw new Exception("Database integrity error");
		return $val === 0;
	}
	
	// Private functions
	private function pushValidationError($field, $message)
	{
		$this->validation[] = array('field'=>$field, 'message'=>$message);
	}
	
	private function flushValidationErrors()
	{
		if (count($this->validation))
			throw new ValidationError($this->validation);
	}
	
	private $validation;
	private $app;
};
