#!/usr/bin/python
# -*- coding: utf-8 -*-

from nijmetrics import Application, sanitizefn
import cPickle as pickle
import sys
import itertools
import zlib
import copy
import types


'''
Calculate the coherency metrics, named M1<x> where x is a,b,c,d,e, M4 and M2<x,y>
The description of each is noted below. 
-> comment: denotes a modification to the specification that this script 
   implements (additions only, so they do not hurt)

M4: # of minutiae of each type (ridge endings, bifurcations, type unknowns, position unknowns)

M1a: # of minutiae placed outside of all quality zones
M1b: (#re+#bif)/(#total_in_zone) for the high quality (q=1) zone and low quality (q=3) zones
	-> added orange zone as well
M1c: # of minutiae in green area / surface of green area
	-> calculated for all 3 areas
M1d: (#re+#bif)/#total
M1e: surface of green area / total work surface (=total surface with quality areas, != total surface)
	-> calculated for all 3 areas
	
M2a: (# of green pixels) / (surface of mask)
M2b: (a.(# of green pixels) + b.(# of orange pixels)) / (surface of mask) with a+b=1 and a>b (e.g. a=0.7, b=0.3)
    -> M2a and M2b are calculated for all values of mask thresholds
	
M3: user's quality maps vs the consensus (more details to come)
'''

class Coherency(Application):

	def __init__(self, filename):
		Application.__init__(self)
		self.filename = filename
		self.metric_desc = {
			'M1a': '# of minutiae placed outside of all quality zones',
			'M1b': ('', 'quality'),
			'M1c': ('', 'quality'),
			'M1d': '',
			'M1e': ('', 'quality'),
			'M2a': ('', 'threshold'),
			'M2b': ('', 'threshold'),
			'M3': ('', 'threshold'),
			'M4': ('', 'mintype'),
		}
		self.metric_names = sorted(self.metric_desc.keys())
		self.situations = [
			'analysis_mark',
			'comparison_mark',
			'comparison_print',
		]
		self.standardTable = {
			1: {'id': 1, 'color': '#2ca02c', 'label': 'High'},
			2: {'id': 2, 'color': '#ff7f0e', 'label': 'Medium'},
			3: {'id': 3, 'color': '#d62728', 'label': 'Low'},
		}
		
		self.m3matrix = {
			(1,1): 0, (1,2): 1, (1,3): 2, (1,0): 3, 
			(2,1): 1, (2,2): 0, (2,3): 1, (2,0): 2, 
			(3,1): 2, (3,2): 1, (3,3): 0, (3,0): 1, 
			(0,1): 3, (0,2): 2, (0,3): 1, (0,0): 0, 
		}
		
		self.mergedMapColorScale = {
			1:(0,255,0,255), # high (q=1)      -> green
			2:(0,0,255,255), # medium (q=2)    -> blue (!)
			3:(255,0,0,255), # low (q=3)       -> red
			0:(0,0,0,255),   # nothing         -> black
			-1:(128,128,128,255), # masked out -> gray
			-2:(255,0,255,255), # error        -> shitty color (happens when a single pixel is covered by multiple quality level areas)
		}

		self.all_minutia_types = self.getMinutiaTypes(noext=True)
		self.all_qualities = [1,2,3]

	
	def prepare_output_format(self, thresholds):
		self.thresholds = sorted(thresholds)
		n = {}
		for mn,m in self.metric_desc.items():
			if type(m) is type(''):
				m = (m,'single')
			if m[1] == 'single':
				m = [x for x in m] + [['']]
			elif m[1] == 'quality':
				m = [x for x in m] + [self.all_qualities]
			elif m[1] == 'threshold':
				m = [x for x in m] + [self.thresholds]
			elif m[1] == 'mintype':
				m = [x for x in m] + [self.all_minutia_types]
			n[mn] = m
		self.metric_desc = n

		self.header1 = ['General info','','']
		self.header2 = ['Exercise','User','Situation']
		for mn in self.metric_names:
			self.header1 += [mn]
			m = self.metric_desc[mn]
			assert m[2]
			first = True
			for extcol in m[2]:
				if not first:
					self.header1 += ['']
				self.header2 += [extcol]
				
				first = False
			
	def run(self):
	
		with open(self.filename,'r') as f:
			data = pickle.load(f)
		
			
		
		for exid in data:
			outname = sanitizefn('%s'%(data[exid]['exercise']['name']))+'.csv'
		
			self.prepare_output_format(data[exid]['consensus']['masks'].keys())
			
			with open(outname, 'w') as f:
				# Write the column headers
				f.write(';'.join(['"'+str(x)+'"' if type(x) is types.StringType else str(x) for x in self.header1])+'\n')
				f.write(';'.join(['"'+str(x)+'"' if type(x) is types.StringType else str(x) for x in self.header2])+'\n')
				f.flush()
				
				self.metric_names
				self.do_run(exid, data, f)
			
		return True
	
	
	def do_run(self, exid, data, outputf):
	
		
		print "Exercise %s (ID=%d)" % (data[exid]['exercise']['name'],exid)
		
		image = data[exid]['mark']
		#print "  Image: " % ()
		
		ex_diagnostics = {
			'warn_consensus_sum_differs_from_numcontributors': 0,
		}
		
		
		print "  Gathering consensus data ..."
		
		masks = {}
		qmaps_orig = {}
		qmaps = {}
		# Read the quality maps
		print "    Reading quality maps",
		for quality in self.all_qualities:
			print quality,
			qmap = data[exid]['consensus']['qmaps'][quality]
			if data[exid]['compressed']:
				qmap = eval(zlib.decompress(qmap))
			qmaps_orig[quality] = qmap
		print "Done"
		
		# Merge them into a single buffer, where each pixel contains 4 elements:
		# 0 the number of users that did *not* cover the pixel
		# 1,2 and 3: the number of users that covered the pixel in the said quality level
		# Initially, all indices 0 contain the number of contributors. This is subtracted by
		# the number of users who covered in green, orange and red. What remains is the number
		# of users that did not put any quality zone here.
		print "    Merging quality maps",
		qmaps_all = [[copy.copy(data[exid]['consensus']['numcontributors']),0,0,0] for x in range(image['width']*image['height'])]
		for quality in self.all_qualities:
			print quality,
			qmap = qmaps_orig[quality]
			for i in range(len(qmaps_all)):
				num = qmap[i]
				qmaps_all[i][quality] = num
				qmaps_all[i][0] -= num
		for i in range(len(qmaps_all)):
			if qmaps_all[i][0] < 0:
				self.warn(ex_diagnostics, 'warn_consensus_sum_differs_from_numcontributors', "WARNING: at index %d, values are %s (numc=%d)" % (i, str(qmaps_all[i]), data[exid]['consensus']['numcontributors']))
		#self.saveQualityMap(qmaps_all, (image['width'],image['height']), 'consensus_all.png', self.mergedMapColorScale)
		print "Done"
				
		# Read the masks and apply them to the merged qmaps
		print "    Reading and applying masks",
		for threshold,mask in data[exid]['consensus']['masks'].iteritems():
			print threshold,
			if data[exid]['compressed']:
				mask = eval(zlib.decompress(mask))
			masks[threshold] = {'data':mask, 'surface':sum(mask)}

			qm = [0]*(image['width']*image['height'])
			assert len(mask) == len(qm)
			for i in range(len(qm)):
				qm[i] = qmaps_all[i] if mask[i] else -1
			qmaps[threshold] = qm
			#self.saveQualityMap(qm, (image['width'],image['height']), 'consensus_all_masked_%d.png'%(int(threshold*100)), self.mergedMapColorScale)
		print "Done"
		
		count = 0
		for userid in data[exid]['users']:
			print "  User %d  (%.2f%%)" % (userid,float(count)/float(len(data[exid]['users']))*100.0)
			count += 1
			
			userdata = data[exid]['users'][userid]
			
			maps = {}
			print "    Reading quality maps ...",
			for quality in userdata['qmaps']:
				print quality,
				map = userdata['qmaps'][quality]
				if data[exid]['compressed']:
					map = eval(zlib.decompress(map))
				maps[quality] = map
			print "OK"
			
			
			print "    Calculating metrics"
			
				
			# Calculate all metrics for the 3 situations (if posssible): during analysis (mark only)
			# and during comparison (mark and print)
			for situation in self.situations:
				phase,imgtype = situation.split('_')

				# Get a reference to the corresponding image
				assert imgtype in data[exid]
				image = data[exid][imgtype]
				mapdims = (image['width'],image['height'])
				
				# Prepare the storage for the metric values and troubleshooting indicators
				metrics = {}
				diagnostics = {
					'warn_minutiae_in_multiple_areas': 0,
					'warn_union_of_surfaces_differ_from_sum_by': 0,
					'warn_pixel_covered_multiple_times': 0,
				}
				
				# The calculation only happens if there is something to work on (i.e. if the user
				# did submit data for that situation=phase/image)
				if phase in userdata and imgtype in userdata[phase]:
					
					print "      In",phase,"/",imgtype,"...",
					annotations = userdata[phase][imgtype]
					
					# Strip off centers and differences - as these are not actual minutiae in the sense
					# of our analysis right now
					minutiae = {}
					try:
						for id,m in annotations['minutiae'].iteritems():
							if m['type'] not in ('center','difference'):
								minutiae[id] = m
					except AttributeError,err:
						print "WARNING:",err
						print "NOTE: the offending thing that should be a dict of minutiae and isn't is:"
						print repr(annotations['minutiae'])
						
					# M4: # of minutiae of each type (ridge endings, bifurcations, type unknowns, position unknowns)
					print "M4",
					M4 = dict(zip(self.all_minutia_types, [0]*len(self.all_minutia_types)))
					for id,m in minutiae.iteritems():
						M4[m['type']] += 1
					metrics['M4'] = M4
					
					# M1x metrics are calculated on the mark only (because they depend on the quality zones, which
					# are only processed on the mark even though the user could have annotated the quality on the print)
					if imgtype != 'print':
						
						# M1a: # of minutiae placed outside of all quality zones. 
						# Keep a record of the zone the minutia belongs to by the way
						print "M1a",
						mzones = {}
						M1a = 0
						for id,m in minutiae.iteritems():
							inside = 0
							for quality,map in maps.iteritems():
								if self.isPointInMap(m['pos'], map, mapdims):
									inside += 1
									mzones[id] = quality
							if inside == 0:
								M1a += 1
							elif inside != 1:
								self.warn(diagnostics, 'warn_minutiae_in_multiple_areas', "WARNING: in %s minutia %s is inside %d maps" % (situation,id,inside))
						for id,m in minutiae.iteritems():
							if id not in mzones:
								mzones[id] = -1
								
						# M1b: (#re+#bif)/(#total_in_zone) for the high quality (q=1) zone and low quality (q=3) zones
						print "M1b",
						M1b = {}
						for quality in self.all_qualities:
							numre = 0
							numbif = 0
							numtotal = 0
							for id,m in minutiae.iteritems():
								if mzones[id] == quality:
									numtotal += 1
									if m['type'] == 'bifurcation':
										numbif += 1
									elif m['type'] == 'ridge_ending':
										numre += 1
							if numtotal:
								M1b[quality] = float(numre+numbif) / float(numtotal)
							else:
								M1b[quality] = None
								
						# M1c: # of minutiae in green area / surface of green area
						print "M1c",
						surfaces = {}
						M1c = {}
						for quality in self.all_qualities:
							if quality in maps:
								surfaces[quality] = self.getMapSurface(maps[quality])
								if surfaces[quality] > 0:
									nummin = 0
									for id,m in minutiae.iteritems():
										if mzones[id] == quality:
											nummin += 1
									M1c[quality] = float(nummin) / float(surfaces[quality])
								else:
									surfaces[quality] = None
									M1c[quality] = None
							else:
								surfaces[quality] = None
								M1c[quality] = None
								
								
						# M1d: (#re+#bif)/#total
						print "M1d",
						numtotalmin = sum(M4.values())
						if numtotalmin:
							M1d = float(M4['ridge_ending']+M4['bifurcation'])/float(numtotalmin)
						else:
							M1d = None
						
						
						# M1e: surface of green area / total work surface (=total surface with quality areas, != total surface)
						print "M1e",
						# if total area is considered (as before fix of 2012-10-08): totalsurf=float(image['width']*image['height'])
						# now (after fix): calculate union of the 3 quality maps and get its surface
						# Note that it equals the sum of surfaces[x] if and only if all quality layers have no common pixels - which
						# should be the case at this point (i.e. after extraction, and provided it did its job correctly)
						# So ... the sum of surfaces and the surface of the union should in fact be the same. Otherwise there
						# is something to investigate.
						worksurf = [0]*(image['width']*image['height'])
						for quality in self.all_qualities:
							if quality in maps:
								map = maps[quality]
								assert len(map) == len(worksurf)
								for i in range(len(map)):
									worksurf[i] |= map[i]
							assert sum([0 if x in (0,1) else 1 for x in worksurf]) == 0 # There should be only 0s or 1s in the union
						totalsurf = self.getMapSurface(worksurf)
						thesum = sum([surfaces[x] if surfaces[x] is not None else 0 for x in self.all_qualities])
						if totalsurf != thesum:
							print "WARNING: surface of union of quality maps (%d) differs from sum of surfaces (%d). It means %d pixels are covered by more than one quality area." % (totalsurf, thesum, thesum-totalsurf) # If the difference is negative you live in a parallel universe
							diagnostics['warn_union_of_surfaces_differ_from_sum_by'] = thesum-totalsurf
						M1e = {}
						for quality in self.all_qualities:
							if totalsurf > 0.0 and quality in surfaces and surfaces[quality] is not None:
								M1e[quality] = float(surfaces[quality]) / totalsurf
							else:
								M1e[quality] = None
					
						metrics['M1a'] = M1a
						metrics['M1b'] = M1b
						metrics['M1c'] = M1c
						metrics['M1d'] = M1d
						metrics['M1e'] = M1e
						
						
						# M2a: (# of green pixels) / (surface of mask)
						print "M2a",
						M2a = {}
						for threshold,mask in masks.iteritems():
							if surfaces[1] is not None:
								M2a[threshold] = surfaces[1] / mask['surface']
							else:
								M2a[threshold] = None
								
						# M2b: (a.(# of green pixels) + b.(# of orange pixels)) / (surface of mask) with a+b=1 and a>b (e.g. a=0.7, b=0.3)
						print "M2b",
						M2b = {}
						a,b = 0.7, 0.3
						for threshold,mask in masks.iteritems():
							if surfaces[1] is not None and surfaces[2] is not None:
								M2b[threshold] = (a*surfaces[1] + b*surfaces[2]) / mask['surface']
							else:
								M2b[threshold] = None
				
						metrics['M2a'] = M2a
						metrics['M2b'] = M2b
						
					print "Done"
					
				
				if situation == 'analysis_mark':
					print "      QMap vs. consensus",
					# M3: user's quality maps vs the consensus
					M3 = {}
					# Get the user's full map (union of all three layers)
					fullmap = [0]*(image['width']*image['height'])
					for quality in self.all_qualities:
						if quality in maps:
							map = maps[quality]
							#self.saveQualityMap(map, (image['width'],image['height']), 'qmap_q%d.png'%quality)
							assert len(map) == len(fullmap)
							for i in range(len(map)):
								if fullmap[i] == 0:
									fullmap[i] = quality if map[i] else 0
								elif map[i] != 0:
									fullmap[i] = -2
									self.warn(diagnostics, 'warn_pixel_covered_multiple_times', "WARNING: pixel %d (x=%d,y=%d) was assigned quality %d, but %d applies as well - keeping old" % (i, i%image['width'], i/image['width'], fullmap[i], quality))
							#self.saveQualityMap(fullmap, (image['width'],image['height']), sanitizefn('%s_%d_qmap_all'%(data[exid]['exercise']['name'], userid))+'.png', self.mergedMapColorScale)
					
					for threshold,maskdef in masks.iteritems():
						print threshold,
						# Do not apply the mask to the user's: we already discard pixels because the consensus qmap has been filtered already
						fm = fullmap
						# Apply the mask to the user's map
						# fm = [0]*(image['width']*image['height'])
						# mask = maskdef['data']
						# assert len(mask) == len(fm)
						# for i in range(len(fm)):
							# fm[i] = fullmap[i] if mask[i] else -1
						#self.saveQualityMap(fm, (image['width'],image['height']), sanitizefn('%s_%d_qmap_all_masked_%d'%(data[exid]['exercise']['name'], userid,int(threshold*100)))+'.png',self.mergedMapColorScale)
					
						# Calculate using the matrix
						fnumc = float(data[exid]['consensus']['numcontributors'])
						badvalues = (-1,-2)
						goodvalues = (0,1,2,3)
						qm = qmaps[threshold]
						assert len(qm) == len(fm)
						m3 = 0
						for i in range(len(qm)):
							cons = qm[i]
							if cons in badvalues or fm[i] in badvalues:
								continue
							#print "px",i
							for qcoef in goodvalues:
								coef = self.m3matrix[(qcoef, fm[i])]
								prop = min(max(cons[qcoef] / fnumc, 0.0), 1.0)
								#print qcoef,fm[i],coef,prop,coef*prop,cons
								m3 += coef * prop
						
						M3[threshold] = m3
					
					metrics['M3'] = M3
					
					print "Done"
			
				line = [data[exid]['exercise']['name'], userid, situation]
				for mname in self.metric_names:
					m = self.metric_desc[mname]
					# Fetch the python value of the metric
					if mname in metrics:
						mv = metrics[mname]
					else:
						mv = None
					#print '%s=%s'%(mname,str(mv))
					# Write this in the csv, using multiple columns if required
					if m[1] == 'single':
						line.append(mv)
					elif mv is None:
						for item in m[2]:
							line.append(None)
					else:
						#print m[2],mv
						for item in m[2]:
							assert item in mv
							line.append(mv[item])
				assert len(line) == len(self.header2)
				outputf.write(';'.join(['"'+str(x)+'"' if type(x) is types.StringType else str(x) for x in line])+'\n')	
				outputf.flush()
					
	
	# Returns True if the provided point=(x,y) is in the specified map whose dimensions are mapdims=(w,h)
	# The coordinates of the point are rounded to the closest integer
	def isPointInMap(self, point, map, mapdims):
		ipoint = [int(round(x)) for x in point]
		if ipoint[0] < 0 or ipoint[0] >= mapdims[0] or ipoint[1] < 0 or ipoint[1] >= mapdims[1]:
			return False
		index = mapdims[0] * ipoint[1] + ipoint[0]
		# if index >= len(map):
			# print "assert index < len(map) is bound to fail, dumping objects"
			# print "point",point
			# print "mapdims",mapdims
			# print "index",index
			# print "length of map",len(map)
		assert index < len(map)
		assert map[index] == 1 or map[index] == 0
		return map[index] != 0
		
	# Get the surface, in number of pixels, of the provided map
	def getMapSurface(self, map):
		return sum(map)
		
		
	def warn(self, diags, key, message):
		diags[key] += 1
		if diags[key] <= 100:
			if diags[key] == 100:
				print "NOTE: skipping warnings of the same type following next one"
			print message
	